'use strict';
/*global require, module, Buffer, jsGen*/
var client = {},
    redis = require('redis'),
    then = jsGen.module.then,
    redisPort = jsGen.conf.RedisPort || 6379,
    redisIp = jsGen.conf.RedisIp || '127.0.0.1',
    globalCacheDb = jsGen.conf.RedisDefaultDb || 1,
    onlineUserCacheDb = globalCacheDb + 2,
    onlineCacheDb = globalCacheDb + 1,
    TimeLimitCacheId = 0,
    MD5 = jsGen.lib.tools.MD5,
    each = jsGen.lib.tools.each,
    union = jsGen.lib.tools.union,
    toNum = jsGen.lib.tools.toNum,
    throttle = jsGen.lib.tools.throttle,
    intersect = jsGen.lib.tools.intersect,
    callbackFn = jsGen.lib.tools.callbackFn,
    errorHandler = jsGen.lib.tools.errorHandler,
    clientSub = redis.createClient(redisPort, redisIp, {
        connect_timeout: 5000
    }); // 订阅频道等候消息的client

function connect(callback) {
    console.log('Connect to Redis...');
    then.each([globalCacheDb, onlineCacheDb, onlineUserCacheDb], function (cont, x) {
        client[x] = redis.createClient(redisPort, redisIp, {
            connect_timeout: 5000
        }).on('ready', function () {
            client[x].select(x, cont);
        }).on('error', function (error) {
            cont(error);
        });
    }).fin(function (cont, error) {
        if (!error) {
            console.log('Redis connected!');
        }
        callback(error);
    });
}

function close() {
    each(client, function (cli) {
        cli.end();
    });
}

function initConfig(configTpl, callback) {
    var config = {}, // 新构建的config缓存
        _config = union(configTpl), // 从configTpl克隆的config闭包镜像
        subPubId = MD5('' + Date.now() + Math.random(), 'base64'); // 本进程的唯一识别ID

    callback = callback || callbackFn;

    var update = throttle(function () {
        jsGen.dao.index.setGlobalConfig(_config);
    }, 120000); // 将config写入MongoDB，每两分钟内最多执行一次

    function updateKey(key) { // 更新镜像的key键值
        return then(function (cont) {
            client[globalCacheDb].hget('config.hash', key, cont); // 从redis读取更新的数据
        }).then(function (cont, reply) {
            reply = JSON.parse(reply);
            _config[key] = typeof _config[key] === typeof reply ? reply : _config[key]; // 数据写入config镜像
            cont(null, _config[key]);
        });
    }

    clientSub.on('message', function (channel, key) {
        var ID = key.slice(0, 24); // 分离识别ID和key
        key = key.slice(24);
        if (channel === 'updateConfig' && ID !== subPubId) { // 来自于updateConfig频道且不是本进程发出的更新通知
            if (key in _config) {
                updateKey(key); // 更新一个key
            } else {
                each(_config, function (value, key) { // 更新整个config镜像
                    updateKey(key);
                });
            }
        }
    });
    clientSub.subscribe('updateConfig'); // 订阅updateConfig频道

    each(configTpl, function (value, key) { // 从configTpl模板构建getter/setter，利用Hash类型存储config
        Object.defineProperty(config, key, {
            set: function (value) {
                var byStep = (value === 1 || value === -1) && typeof _config[key] === 'number';
                _config[key] = byStep ? _config[key] + value : value;
                then(function (cont) {
                    if (byStep) {
                        // 按1递增或递减，更新镜像，再更新redis
                        client[globalCacheDb].hincrby('config.hash', key, value, cont);
                    } else {
                        // 因为redis存储字符串，下面先序列化。
                        client[globalCacheDb].hset('config.hash', key, JSON.stringify(value), cont);
                    }
                }).then(function () { // redis数据库更新完成，向其他进程发出更新通知
                    client[globalCacheDb].publish('updateConfig', subPubId + key);
                }).fail(jsGen.thenErrLog);
                update(); // 更新MongoDB
            },
            get: function () {
                return _config[key]; // 从镜像读取数据
            },
            enumerable: true,
            configurable: true
        });
    });
    callback(null, config);
    return config; // 同步返回新的config对象
}

function TimeLimitCache(timeLimit, type, name, getUpdate) {
    var method = {
        string: ['get', 'set'],
        array: ['lrange', 'lpush'],
        object: ['hgetall', 'hmset']
    };
    this.name = 'tlCache.' + (name || TimeLimitCacheId++) + '.';
    this.method = method[type || 'string'];
    this.timeLimit = +timeLimit || 60;
    this.getUpdate = !! getUpdate;
}

TimeLimitCache.prototype.get = function (key, callback) {
    var that = this;
    callback = callback || callbackFn;
    then(function (cont) {
        var get = that.method[0];
        if (that.getUpdate) {
            client[globalCacheDb].expire(that.name + key, that.timeLimit);
        }
        if (get === 'lrange') {
            client[globalCacheDb][get](that.name + key, 0, -1, callback);
        } else {
            client[globalCacheDb][get](that.name + key, callback);
        }
    }).fail(function (cont, err) {
        callback(err);
    });
};
TimeLimitCache.prototype.put = function (key, value, callback) {
    var that = this;
    callback = callback || callbackFn;
    then(function (cont) {
        client[globalCacheDb].del(that.name + key, cont);
    }).then(function (cont) {
        client[globalCacheDb][that.method[1]](that.name + key, value, cont);
    }).then(function (cont) {
        client[globalCacheDb].expire(that.name + key, that.timeLimit, callback);
    }).fail(function (cont, err) {
        callback(err);
    });
};
TimeLimitCache.prototype.remove = function (key, callback) {
    var that = this;
    callback = callback || callbackFn;
    then(function (cont) {
        client[globalCacheDb].del(that.name + key, callback);
    }).fail(function (cont, err) {
        callback(err);
    });
};
TimeLimitCache.prototype.removeAll = function (callback) {
    var that = this;
    callback = callback || callbackFn;
    then(function (cont) {
        client[globalCacheDb].keys(that.name + '*', cont);
    }).then(function (cont, keys) {
        if (keys.length > 0) {
            client[globalCacheDb].del(keys, cont);
        } else {
            cont(null, 0);
        }
    }).then(function (cont, num) {
        callback(null, num);
    }).fail(function (cont, err) {
        callback(err);
    });
};
TimeLimitCache.prototype.info = function (callback) {
    var that = this;
    callback = callback || callbackFn;
    then(function (cont) {
        client[globalCacheDb].keys(that.name + '*', cont);
    }).then(function (cont, keys) {
        keys = keys || [];
        callback(null, {
            name: that.name,
            timeLimit: that.timeLimit,
            length: keys.length
        });
    }).fail(function (cont, err) {
        callback(err);
    });
};

function onlineCache(req, callback) {
    var users = 0,
        key = req.session.Uid || MD5(req.useragent + req.ip, 'base64');

    callback = callback || callbackFn;
    if (key) {
        then(function (cont) {
            client[req.session.Uid ? onlineUserCacheDb : onlineCacheDb].setex(key, 600, '', cont);
        }).then(function (cont) {
            client[onlineUserCacheDb].dbsize(cont);
        }).then(function (cont, reply) {
            users = +reply;
            client[onlineCacheDb].dbsize(cont);
        }).then(function (cont, reply) {
            callback(null, users, +reply + users);
        }).fail(function (cont, err) {
            callback(err);
        });
    } else {
        callback(jsGen.Err('redis onlineCache request error!'));
    }
}

function userCache(_id, callback) {
    callback = callback || callbackFn;

    if (_id > 0) {
        then(function (cont) {
            client[globalCacheDb].hgetall('user.hash.' + _id, cont);
        }).then(function (cont, user) {
            if (!user) {
                callback(jsGen.Err('redis user not exist!'));
            } else {
                user._id = +user._id;
                callback(null, user);
            }
        }).fail(function (cont, err) {
            callback(err);
        });
    } else {
        callback(jsGen.Err('redis userCache request error!'));
    }
}
userCache.index = function (start, end, callback) {
    start = start || 0;
    end = end || 50;
    callback = callback || callbackFn;
    then(function (cont) {
        client[globalCacheDb].zrevrange('user.sset.index', start, end, cont);
    }).then(function (cont, list) {
        callback(null, toNum(list || []));
    }).fail(function (cont, err) {
        callback(err);
    });
};
userCache.index.total = function (callback) {
    callback = callback || callbackFn;
    then(function (cont) {
        client[globalCacheDb].zcard('user.sset.index', cont);
    }).then(function (cont, total) {
        callback(null, +total);
    }).fail(function (cont, err) {
        callback(err);
    });
};
userCache.get = function (name, callback) {
    callback = callback || callbackFn;

    if (name) {
        then(function (cont) {
            client[globalCacheDb].get(('user.str.' + name).toLowerCase(), cont);
        }).then(function (cont, _id) {
            callback(_id > 0 ? null : jsGen.Err('redis user not exist!'), +_id);
        }).fail(function (cont, err) {
            callback(err);
        });
    } else {
        callback(jsGen.Err('redis userCache.get request error!'));
    }
};
userCache.update = function (userObj, callback) {
    var user = {
        _id: 0,
        name: '',
        email: '',
        avatar: '',
        score: 0,
        role: 0
    };

    callback = callback || callbackFn;
    intersect(user, userObj);
    if (user._id > 0) {
        then(function (cont) {
            client[globalCacheDb].hgetall('user.hash.' + user._id, cont);
        }).fin(function (cont, err, old) {
            if (old) {
                then(function (cont2) {
                    if (user.name && old.name !== user.name) {
                        client[globalCacheDb].del(('user.str.' + old.name).toLowerCase(), cont2);
                    } else {
                        cont2();
                    }
                }).then(function (cont2) {
                    if (user.email && old.email !== user.email) {
                        client[globalCacheDb].del(('user.str.' + old.email).toLowerCase(), cont);
                    } else {
                        cont();
                    }
                }).fail(function (cont2, err) {
                    cont();
                });
            } else {
                cont();
            }
        }).then(function (cont) {
            client[globalCacheDb].hmset('user.hash.' + user._id, user, cont);
        }).then(function (cont) {
            client[globalCacheDb].zadd('user.sset.index', user._id, user._id, cont);
        }).then(function (cont) {
            client[globalCacheDb].set(('user.str.' + user.name).toLowerCase(), user._id, cont);
        }).then(function (cont) {
            client[globalCacheDb].set(('user.str.' + user.email).toLowerCase(), user._id, cont);
        }).then(function (cont) {
            callback(null, true);
        }).fail(function (cont, err) {
            callback(err);
        });
    } else {
        callback(jsGen.Err('redis userCache update request error!'));
    }
};
userCache.removeAll = function (callback) {
    callback = callback || callbackFn;
    then(function (cont) {
        client[globalCacheDb].keys('user*', cont);
    }).then(function (cont, keys) {
        if (keys.length > 0) {
            client[globalCacheDb].del(keys, cont);
        } else {
            cont(null, 0);
        }
    }).then(function (cont, num) {
        callback(null, num);
    }).fail(function (cont, err) {
        callback(err);
    });
};

function tagCache(_id, callback) {
    callback = callback || callbackFn;

    if (_id > 0) {
        then(function (cont) {
            client[globalCacheDb].hgetall('tag.hash.' + _id, cont);
        }).then(function (cont, tag) {
            if (!tag) {
                callback(jsGen.Err('redis tag not exist!'));
            } else {
                tag._id = +tag._id;
                tag.articles = +tag.articles;
                tag.users = +tag.users;
                callback(null, tag);
            }
        }).fail(function (cont, err) {
            callback(err);
        });
    } else {
        callback(jsGen.Err('redis tagCache request error!'));
    }
}
tagCache.index = function (start, end, callback) {
    start = start || 0;
    end = end > 1 ? end - 1 : -1;
    callback = callback || callbackFn;
    then(function (cont) {
        client[globalCacheDb].zrevrange('tag.sset.index', start, end, cont);
    }).then(function (cont, list) {
        callback(null, toNum(list || []));
    }).fail(function (cont, err) {
        callback(err);
    });
};
tagCache.index.total = function (callback) {
    callback = callback || callbackFn;
    then(function (cont) {
        client[globalCacheDb].zcard('tag.sset.index', cont);
    }).then(function (cont, total) {
        callback(null, +total);
    }).fail(function (cont, err) {
        callback(err);
    });
};
tagCache.get = function (name, callback) {
    callback = callback || callbackFn;

    if (name) {
        then(function (cont) {
            client[globalCacheDb].get(('tag.str.' + name).toLowerCase(), cont);
        }).then(function (cont, _id) {
            callback(_id > 0 ? null : jsGen.Err('redis tag not exist!'), +_id);
        }).fail(function (cont, err) {
            callback(err);
        });
    } else {
        callback(jsGen.Err('redis tagCache.get request error!'));
    }
};
tagCache.update = function (tagObj, callback) {
    var tag = {
        _id: 0,
        tag: '',
        articles: 0,
        users: 0
    };

    callback = callback || callbackFn;
    intersect(tag, tagObj);
    if (tag._id > 0) {
        then(function (cont) {
            client[globalCacheDb].hgetall('tag.hash.' + tag._id, cont);
        }).fin(function (cont, err, old) {
            if (old && tag.tag && old.tag !== tag.tag) {
                client[globalCacheDb].del(('tag.str.' + old.tag).toLowerCase(), cont);
            } else {
                cont();
            }
        }).then(function (cont) {
            client[globalCacheDb].hmset('tag.hash.' + tag._id, tag, cont);
        }).then(function (cont) {
            client[globalCacheDb].zadd('tag.sset.index', tag.articles, tag._id, cont);
        }).then(function (cont) {
            client[globalCacheDb].set(('tag.str.' + tag.tag).toLowerCase(), tag._id, cont);
        }).then(function (cont) {
            callback(null, true);
        }).fail(function (cont, err) {
            callback(err);
        });
    } else {
        callback(jsGen.Err('redis tagCache update request error!'));
    }
};
tagCache.remove = function (_id, callback) {
    callback = callback || callbackFn;
    then(function (cont) {
        client[globalCacheDb].del('tag.hash.' + _id, cont);
    }).then(function (cont) {
        client[globalCacheDb].zrem('tag.sset.index', _id, cont);
    }).then(function (cont) {
        callback(null, true);
    }).fail(function (cont, err) {
        callback(err);
    });
};
tagCache.removeAll = function (callback) {
    callback = callback || callbackFn;
    then(function (cont) {
        client[globalCacheDb].keys('tag*', cont);
    }).then(function (cont, keys) {
        if (keys.length > 0) {
            client[globalCacheDb].del(keys, cont);
        } else {
            cont(null, 0);
        }
    }).then(function (cont, num) {
        callback(null, num);
    }).fail(function (cont, err) {
        callback(err);
    });
};

function articleCache(_id, callback) {
    callback = callback || callbackFn;

    if (_id > 0) {
        then(function (cont) {
            client[globalCacheDb].hgetall('article.hash.' + _id, cont);
        }).then(function (cont, article) {
            if (!article) {
                callback(jsGen.Err('redis article not exist!'));
            } else {
                article._id = +article._id;
                article.author = +article.author;
                article.date = +article.date;
                article.display = +article.display;
                article.status = +article.status;
                article.updateTime = +article.updateTime;
                article.hots = +article.hots;
                callback(null, article);
            }
        }).fail(function (cont, err) {
            callback(err);
        });
    } else {
        callback(jsGen.Err('redis articleCache request error!'));
    }
}
articleCache.index = function (start, end, callback) {
    start = start || 0;
    end = end > 1 ? end - 1 : -1;
    callback = callback || callbackFn;
    then(function (cont) {
        client[globalCacheDb].zrevrange('article.sset.index', start, end, cont);
    }).then(function (cont, list) {
        callback(null, toNum(list || []));
    }).fail(function (cont, err) {
        callback(err);
    });
};
articleCache.index.total = function (callback) {
    callback = callback || callbackFn;
    then(function (cont) {
        client[globalCacheDb].zcard('article.sset.index', cont);
    }).then(function (cont, total) {
        callback(null, +total);
    }).fail(function (cont, err) {
        callback(err);
    });
};
articleCache.updateList = function (start, end, callback) {
    start = start || 0;
    end = end > 1 ? end - 1 : -1;
    callback = callback || callbackFn;
    then(function (cont) {
        client[globalCacheDb].zrevrange('article.sset.updateList', start, end, cont);
    }).then(function (cont, list) {
        callback(null, list || []);
    }).fail(function (cont, err) {
        callback(err);
    });
};
articleCache.hotsList = function (start, end, callback) {
    start = start || 0;
    end = end > 1 ? end - 1 : -1;
    callback = callback || callbackFn;
    then(function (cont) {
        client[globalCacheDb].zrevrange('article.sset.hotsList', start, end, cont);
    }).then(function (cont, list) {
        callback(null, list || []);
    }).fail(function (cont, err) {
        callback(err);
    });
};
articleCache.hotCommentsList = function (start, end, callback) {
    start = start || 0;
    end = end > 1 ? end - 1 : -1;
    callback = callback || callbackFn;
    then(function (cont) {
        client[globalCacheDb].zrevrange('article.sset.hotCommentsList', start, end, cont);
    }).then(function (cont, list) {
        callback(null, list || []);
    }).fail(function (cont, err) {
        callback(err);
    });
};
articleCache.update = function (articleObj, callback) {
    var article = {
        _id: 0,
        author: 0,
        display: 0,
        status: 0,
        updateTime: 0,
        date: 0,
        hots: 0
    };

    callback = callback || callbackFn;
    intersect(article, articleObj);
    if (article._id > 0) {
        then(function (cont) {
            client[globalCacheDb].hmset('article.hash.' + article._id, article, cont);
        }).then(function (cont) {
            if (article.display === 0 && article.status > -1) {
                then(function (cont2) {
                    //按提交日期排序的文章列表
                    client[globalCacheDb].zadd('article.sset.index', article.status === 2 ? (Date.now() + 604800000) : article.date, article._id, cont2);
                }).then(function (cont2) {
                    //按更新日期排序的文章列表
                    client[globalCacheDb].zadd('article.sset.updateList', article.updateTime, article._id, cont2);
                }).then(function (cont2) {
                    //按热度排序的文章列表
                    client[globalCacheDb].zadd('article.sset.hotsList', article.hots, article._id, cont);
                }).fail(cont);
            } else if (article.status === -1) {
                //按热度排序的评论列表
                client[globalCacheDb].zadd('article.sset.hotCommentsList', article.hots, article._id, cont);
            } else if (article.display > 0) {
                //删除文章
                client[globalCacheDb].zrem('article.sset.index', article._id);
                client[globalCacheDb].zrem('article.sset.updateList', article._id);
                client[globalCacheDb].zrem('article.sset.hotsList', article._id);
                cont();
            }
        }).then(function (cont) {
            callback(null, true);
        }).fail(function (cont, err) {
            callback(err);
        });
    } else {
        callback(jsGen.Err('redis articleCache update request error!'));
    }
};

articleCache.clearup = throttle(function () {
    var now = Date.now();

    then(function (cont) {
        // 移除三个月前的更新
        client[globalCacheDb].zremrangebyscore('article.sset.updateList', 0, now - 7776000000, cont);
    }).then(function (cont) {
        // 获取更新排序列表
        client[globalCacheDb].zrevrange('article.sset.updateList', 0, -1, cont);
    }).then(function (cont, list) {
        list = list || [];
        then(function (cont2) {
            client[globalCacheDb].zrevrange('article.sset.hotsList', 0, -1, cont2);
        }).then(function (cont2, hotsList) {
            //从hotsList移除三月前更新的文章，只保留三个月内更新的文章
            var outDateList = ['article.sset.hotsList'];

            each(hotsList || [], function (x) {
                if (list.indexOf(x) === -1) {
                    outDateList.push(x);
                }
            });
            if (outDateList.length > 1) {
                client[globalCacheDb].zrem(outDateList, cont2);
            } else {
                cont2();
            }
        }).then(function (cont2) {
            client[globalCacheDb].zrevrange('article.sset.hotCommentsList', 0, -1, cont2);
        }).then(function (cont2, hotCommentsList) {
            // 从hotCommentsList移除一个月前更新的评论，只保留一个月内更新的评论
            var outDateList = ['article.sset.hotCommentsList'];
            then.each(hotCommentsList, function (next, x) {
                client[globalCacheDb].hgetall('article.hash.' + x, function (err, comment) {
                    if (!comment || now - comment.updateTime > 2592000000) {
                        outDateList.push(x);
                    }
                    return next ? next() : (outDateList.length > 1 ? client[globalCacheDb].zrem(outDateList, cont2) : cont2());
                });
            });
        }).fail(cont);
    });
}, 300000); // 每五分钟内最多执行一次

articleCache.removeAll = function (callback) {
    callback = callback || callbackFn;
    then(function (cont) {
        client[globalCacheDb].keys('article*', cont);
    }).then(function (cont, keys) {
        if (keys.length > 0) {
            client[globalCacheDb].del(keys, cont);
        } else {
            cont(null, 0);
        }
    }).then(function (cont, num) {
        callback(null, num);
    }).fail(function (cont, err) {
        callback(err);
    });
};

module.exports = {
    redis: redis,
    close: close,
    client: client,
    connect: connect,
    tagCache: tagCache,
    userCache: userCache,
    onlineCache: onlineCache,
    articleCache: articleCache,
    TimeLimitCache: TimeLimitCache,
    initConfig: initConfig
};